@@ -38,6 +38,8 @@
 #include "utils.h"
 #include "seccomp.h"
 
+#define MAX_BUF 1000
+
 bool verify_appname(const char *appname) {
    // these chars are allowed in a appname
    const char* whitelist_re = "^[a-z0-9][a-z0-9+._-]+$";
@@ -132,7 +134,7 @@ void setup_devices_cgroup(const char *appname) {
    // extra paranoia
    if(!verify_appname(appname))
       die("appname %s not allowed", appname);
-   
+
    // create devices cgroup controller
    char cgroup_dir[PATH_MAX];
    must_snprintf(cgroup_dir, sizeof(cgroup_dir), "/sys/fs/cgroup/devices/snappy.%s/", appname);
@@ -160,7 +162,7 @@ bool snappy_udev_setup_required(const char *appname) {
    // extra paranoia
    if(!verify_appname(appname))
       die("appname %s not allowed", appname);
-   
+
    char override_file[PATH_MAX];
    must_snprintf(override_file, sizeof(override_file), "/var/lib/apparmor/clicks/%s.json.additional", appname);
 
@@ -194,10 +196,77 @@ bool snappy_udev_setup_required(const char *appname) {
       debug("found needle, need to apply udev setup");
       return true;
    }
-   
+
    return false;
 }
 
+void setup_private_mount(const char* appname) {
+    uid_t uid = getuid();
+    gid_t gid = getgid();
+    char tmpdir[MAX_BUF] = {0};
+
+    must_snprintf(tmpdir, MAX_BUF, "/tmp/snap.%d_%s_XXXXXX", uid, appname);
+    if (mkdtemp(tmpdir) == NULL) {
+        die("unable to create tmpdir");
+    }
+
+    // unshare() and CLONE_NEWNS require linux >= 2.6.16 and glibc >= 2.14
+    // if using an older glibc, you'd need -D_BSD_SOURCE or -D_SVID_SORUCE.
+    if (unshare(CLONE_NEWNS) < 0) {
+        die("unable to set up mount namespace");
+    }
+
+    // MS_PRIVATE needs linux > 2.6.11
+    if (mount("none", "/tmp", NULL, MS_PRIVATE, NULL) != 0) {
+        die("unable to make /tmp/ private");
+    }
+
+    // MS_BIND is there from linux 2.4
+    if (mount(tmpdir, "/tmp", NULL, MS_BIND, NULL) != 0) {
+        die("unable to bind private /tmp");
+    }
+
+    // do the chown after the bind mount to avoid potential shenanigans
+    if (chown("/tmp/", uid, gid) < 0) {
+        die("unable to chown tmpdir");
+    }
+}
+
+// best-effort attempt at creating the old /tmp/snaps/* TMPDIR.
+void mkoldtmpdir() {
+    char *dir = getenv("TMPDIR");
+    if (!dir || !*dir) {
+        // TMPDIR not set, or empty
+        return;
+    }
+
+    if (strncmp(dir, "/tmp/snaps/", strlen("/tmp/snaps/")) != 0) {
+        // TMPDIR is not /tmp/snaps/*
+        return;
+    }
+
+    // strtok trashes its argument, so to avoid changing the TMPDIR variable,
+    // we make a copy.
+    dir = strdup(dir);
+    if (!dir) {
+        die("out of memory");
+    }
+
+    int n = 4;
+    char buf[MAX_BUF] = "/tmp";
+    char *d = strtok(dir+4, "/");
+    while (d) {
+        n += must_snprintf(buf+n, MAX_BUF-n, "/%s", d);
+        if (mkdir(buf, 01777) < 0 && errno != EEXIST) {
+            break;
+        }
+
+        d = strtok(NULL, "/");
+    }
+
+    free(dir);
+}
+
 int main(int argc, char **argv)
 {
    const int NR_ARGS = 3;
@@ -211,14 +280,6 @@ int main(int argc, char **argv)
    if(!verify_appname(appname))
       die("appname %s not allowed", appname);
 
-   // verify binary path
-   char apps_prefix[128];
-   char oem_prefix[128];
-   must_snprintf(apps_prefix, sizeof(apps_prefix), "/apps/%s/", appname);
-   must_snprintf(oem_prefix, sizeof(oem_prefix), "/oem/%s/", appname);
-   if (strstr(binary, apps_prefix) != binary && strstr(binary, oem_prefix) != binary)
-      die("binary must be inside /apps/%s/ or /oem/%s/", appname, appname);
-      
    // this code always needs to run as root for the cgroup/udev setup,
    // however for the tests we allow it to run as non-root
    if(geteuid() != 0 && getenv("UBUNTU_CORE_LAUNCHER_NO_ROOT") == NULL) {
@@ -226,12 +287,28 @@ int main(int argc, char **argv)
    }
 
    if(geteuid() == 0) {
+       // verify binary path
+       char apps_prefix[128];
+       char frameworks_prefix[128];
+       char oem_prefix[128];
+       must_snprintf(apps_prefix, sizeof(apps_prefix), "/apps/%s/", appname);
+       must_snprintf(frameworks_prefix, sizeof(frameworks_prefix), "/frameworks/%s/", appname);
+       must_snprintf(oem_prefix, sizeof(oem_prefix), "/oem/%s/", appname);
+       if (strstr(binary, apps_prefix) != binary &&
+               strstr(binary, oem_prefix) != binary &&
+               strstr(binary, frameworks_prefix) != binary)
+          die("binary must be inside /apps/%s/, /frameworks/%s/ or /oem/%s/",
+                  appname, appname, appname);
+
+       // set up private mounts
+       setup_private_mount(appname);
+
        // this needs to happen as root
        if(snappy_udev_setup_required(appname)) {
           setup_devices_cgroup(appname);
           setup_udev_snappy_assign(appname);
        }
-      
+
        // the rest does not so drop privs back to calling user
        unsigned real_uid = getuid();
        unsigned real_gid = getgid();
@@ -241,7 +318,7 @@ int main(int argc, char **argv)
        if (setgid(real_gid) != 0)
           die("setgid failed");
        if (setuid(real_uid) != 0)
-          die("seteuid failed");
+          die("setuid failed");
 
        if(real_gid != 0 && (getuid() == 0 || geteuid() == 0))
           die("dropping privs did not work");
@@ -249,9 +326,11 @@ int main(int argc, char **argv)
           die("dropping privs did not work");
     }
 
-   //https://wiki.ubuntu.com/SecurityTeam/Specifications/SnappyConfinement#ubuntu-snapp-launch
+    mkoldtmpdir();
 
-   int rc = 0;
+    //https://wiki.ubuntu.com/SecurityTeam/Specifications/SnappyConfinement#ubuntu-snapp-launch
+
+    int rc = 0;
     // set apparmor rules
     rc = aa_change_onexec(aa_profile);
     if (rc != 0) {
